# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from collections import OrderedDict
from typing import Optional

from mnemonic import Mnemonic
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QApplication, QDialog, QWidget
from substrateinterface import Keypair, KeypairType

from tikka.domains.application import Application
from tikka.domains.entities.constants import (
    DATA_PATH,
    MNEMONIC_WORDS_LENGTH,
    WALLETS_PASSWORD_LENGTH,
)
from tikka.libs.secret import generate_alphabetic
from tikka.slots.pyqt.entities.constants import ADDRESS_MONOSPACE_FONT_NAME
from tikka.slots.pyqt.resources.gui.windows.account_create_rc import (
    Ui_AccountCreateDialog,
)


class AccountCreateWindow(QDialog, Ui_AccountCreateDialog):
    """
    AccountCreateWindow class
    """

    language_code_map = {
        "en": "english",
        "fr": "french",
        "zh-hans": "chinese_simplified",
        "zh-hant": "chinese_traditional",
        "it": "italian",
        "ja": "japanese",
        "ko": "korean",
        "es": "spanish",
    }

    def __init__(self, application: Application, parent: Optional[QWidget] = None):
        """
        Init create account window

        :param application: Application instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        # Mnemonic language selector translated
        mnemonic_language_selector = OrderedDict(
            [
                ("en", self._("English")),
                ("fr", self._("French")),
                ("zh-hans", self._("Chinese simplified")),
                ("zh-hant", self._("Chinese traditional")),
                ("it", self._("Italian")),
                ("ja", self._("Japanese")),
                ("ko", self._("Korean")),
                ("es", self._("Spanish")),
            ]
        )
        for language_code, language_name in mnemonic_language_selector.items():
            self.mnemonicLanguageComboBox.addItem(language_name, userData=language_code)
        self.mnemonicLanguageComboBox.setCurrentIndex(
            self.mnemonicLanguageComboBox.findData(
                self.application.config.get("language")[:2]
            )
        )
        self.language_code = self.mnemonicLanguageComboBox.currentData()

        monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(monospace_font)

        # hide all password stuff and button box
        self.passwordFrame.hide()
        self.buttonBox.button(self.buttonBox.Ok).setDisabled(True)

        # events
        self.changeButton.clicked.connect(self._generate_mnemonic_and_address)
        self.mnemonicLanguageComboBox.currentIndexChanged.connect(
            self._on_mnemonic_language_combo_box_index_changed
        )
        self.passwordChangeButton.clicked.connect(self._generate_wallet_password)
        self.buttonBox.accepted.connect(self.on_accepted_button)
        self.buttonBox.rejected.connect(self.close)
        self.saveMnemonicCheckBox.stateChanged.connect(
            self.on_save_mnemonic_checkbox_state_changed
        )

        # fill form
        self._generate_mnemonic_and_address()
        self._generate_wallet_password()

    def _on_mnemonic_language_combo_box_index_changed(self):
        """
        Triggered when user change mnemonic language selector

        :return:
        """
        old_mnemonic = self.mnemonicLineEdit.text()

        # use mnemonic utilities to get entropy from user language
        mnemonic_util = Mnemonic(self.language_code_map[self.language_code])
        entropy = mnemonic_util.to_entropy(mnemonic_util.normalize_string(old_mnemonic))

        self.language_code = self.mnemonicLanguageComboBox.currentData()

        # get mnemonic in new language from entropy
        mnemonic_util = Mnemonic(self.language_code_map[self.language_code])
        mnemonic = mnemonic_util.to_mnemonic(entropy)

        self.mnemonicLineEdit.setText(mnemonic)

    def _generate_mnemonic_and_address(self):
        """
        Generate mnemonic passphrase and address

        :return:
        """
        mnemonic = Keypair.generate_mnemonic(MNEMONIC_WORDS_LENGTH, self.language_code)
        address = Keypair.create_from_uri(
            suri=mnemonic,
            ss58_format=self.application.currencies.get_current().ss58_format,
            crypto_type=KeypairType.SR25519,
            language_code=self.language_code,
        ).ss58_address

        self.mnemonicLineEdit.setText(mnemonic)
        self.addressValueLabel.setText(address)

    def _generate_wallet_password(self):
        """
        Generate new password for wallet encryption in UI

        :return:
        """
        self.passwordLineEdit.setText(generate_alphabetic(WALLETS_PASSWORD_LENGTH))

    def on_save_mnemonic_checkbox_state_changed(self):
        """
        Triggered when user click on save mnemonic remoinder checkbox

        :return:
        """
        self.buttonBox.button(self.buttonBox.Ok).setEnabled(
            self.saveMnemonicCheckBox.isChecked()
        )
        self.passwordFrame.setHidden(not self.saveMnemonicCheckBox.isChecked())

    def on_accepted_button(self):
        """
        Triggered when user click on ok button

        :return:
        """
        mnemonic = self.mnemonicLineEdit.text()
        language_code = self.mnemonicLanguageComboBox.currentData()
        name = self.nameLineEdit.text().strip()
        password = self.passwordLineEdit.text()

        # create derived account + wallet (and the read-only root account)
        self.application.accounts.create_new_account(
            mnemonic, language_code, "", name, password
        )

        # close window
        self.close()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    AccountCreateWindow(application_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from typing import Dict, List, Optional

from tikka.interfaces.adapters.network.accounts import (
    NetworkAccountsException,
    NetworkAccountsInterface,
)
from tikka.interfaces.adapters.network.connection import NetworkConnectionError


class NetworkAccounts(NetworkAccountsInterface):
    """
    NetworkAccounts class
    """

    def get_balance(self, address: str) -> Optional[int]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAccountsInterface.get_balance.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkAccountsException(NetworkConnectionError())

        # system.account: FrameSystemAccountInfo
        # {
        #   nonce: 1
        #   consumers: 0
        #   providers: 1
        #   sufficients: 0
        #   data: {
        #     randomId: 0x18a4d...
        #     free: 9,799
        #     reserved: 0
        #     feeFrozen: 0
        #   }
        # }
        try:
            result = self.connections.rpc.client.query("System", "Account", [address])
        except Exception as exception:
            logging.exception(exception)
            raise NetworkAccountsException(exception)

        if result.meta_info["result_found"] is False:
            balance = None
        else:
            balance = result.value["data"]["free"]

        return balance

    def get_balances(self, addresses: List[str]) -> Dict[str, Optional[int]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkAccountsInterface.get_balances.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkAccountsException(NetworkConnectionError())

        storage_keys = []
        for address in addresses:
            storage_keys.append(
                self.connections.rpc.client.create_storage_key(
                    "System", "Account", [address]
                )
            )

        try:
            multi_result = self.connections.rpc.client.query_multi(storage_keys)
        except Exception as exception:
            logging.exception(exception)
            raise NetworkAccountsException(exception)

        balances: Dict[str, Optional[int]] = {}
        for storage_key, value_obj in multi_result:
            if value_obj.meta_info["result_found"] is False:
                balances[storage_key.params[0]] = None
            else:
                balances[storage_key.params[0]] = value_obj.value["data"]["free"]

        return balances



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/transfer.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TransferDialog(object):
    def setupUi(self, TransferDialog):
        TransferDialog.setObjectName("TransferDialog")
        TransferDialog.resize(865, 256)
        self.gridLayout = QtWidgets.QGridLayout(TransferDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.amountLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.amountLabel.setFont(font)
        self.amountLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.amountLabel.setObjectName("amountLabel")
        self.gridLayout.addWidget(self.amountLabel, 5, 0, 1, 1)
        self.sendButton = QtWidgets.QPushButton(TransferDialog)
        self.sendButton.setObjectName("sendButton")
        self.gridLayout.addWidget(self.sendButton, 6, 3, 1, 1)
        self.senderNameOrAddressLineEdit = QtWidgets.QLineEdit(TransferDialog)
        self.senderNameOrAddressLineEdit.setClearButtonEnabled(True)
        self.senderNameOrAddressLineEdit.setObjectName("senderNameOrAddressLineEdit")
        self.gridLayout.addWidget(self.senderNameOrAddressLineEdit, 0, 1, 1, 4)
        self.recipientLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.recipientLabel.setFont(font)
        self.recipientLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.recipientLabel.setObjectName("recipientLabel")
        self.gridLayout.addWidget(self.recipientLabel, 2, 0, 1, 1)
        self.recipientNameOrAddressLineEdit = QtWidgets.QLineEdit(TransferDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.recipientNameOrAddressLineEdit.sizePolicy().hasHeightForWidth())
        self.recipientNameOrAddressLineEdit.setSizePolicy(sizePolicy)
        self.recipientNameOrAddressLineEdit.setClearButtonEnabled(True)
        self.recipientNameOrAddressLineEdit.setObjectName("recipientNameOrAddressLineEdit")
        self.gridLayout.addWidget(self.recipientNameOrAddressLineEdit, 2, 1, 1, 4)
        self.recipientBalanceValueLabel = QtWidgets.QLabel(TransferDialog)
        self.recipientBalanceValueLabel.setText("")
        self.recipientBalanceValueLabel.setObjectName("recipientBalanceValueLabel")
        self.gridLayout.addWidget(self.recipientBalanceValueLabel, 2, 8, 1, 1)
        self.amountUnitComboBox = QtWidgets.QComboBox(TransferDialog)
        self.amountUnitComboBox.setObjectName("amountUnitComboBox")
        self.gridLayout.addWidget(self.amountUnitComboBox, 5, 2, 1, 1)
        self.amountDoubleSpinBox = QtWidgets.QDoubleSpinBox(TransferDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.amountDoubleSpinBox.sizePolicy().hasHeightForWidth())
        self.amountDoubleSpinBox.setSizePolicy(sizePolicy)
        self.amountDoubleSpinBox.setMaximum(999999999999.99)
        self.amountDoubleSpinBox.setObjectName("amountDoubleSpinBox")
        self.gridLayout.addWidget(self.amountDoubleSpinBox, 5, 1, 1, 1)
        self.feesButton = QtWidgets.QPushButton(TransferDialog)
        self.feesButton.setObjectName("feesButton")
        self.gridLayout.addWidget(self.feesButton, 5, 3, 1, 1)
        self.senderLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.senderLabel.setFont(font)
        self.senderLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.senderLabel.setObjectName("senderLabel")
        self.gridLayout.addWidget(self.senderLabel, 0, 0, 1, 1)
        self.recipientNameOrAddressValueLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.recipientNameOrAddressValueLabel.setFont(font)
        self.recipientNameOrAddressValueLabel.setText("")
        self.recipientNameOrAddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.recipientNameOrAddressValueLabel.setObjectName("recipientNameOrAddressValueLabel")
        self.gridLayout.addWidget(self.recipientNameOrAddressValueLabel, 3, 1, 1, 4)
        self.errorLabel = QtWidgets.QLabel(TransferDialog)
        self.errorLabel.setStyleSheet("")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.gridLayout.addWidget(self.errorLabel, 7, 1, 1, 4)
        self.buttonBox = QtWidgets.QDialogButtonBox(TransferDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonBox.sizePolicy().hasHeightForWidth())
        self.buttonBox.setSizePolicy(sizePolicy)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 7, 8, 1, 1, QtCore.Qt.AlignHCenter)
        self.feesLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.feesLabel.setFont(font)
        self.feesLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.feesLabel.setObjectName("feesLabel")
        self.gridLayout.addWidget(self.feesLabel, 6, 0, 1, 1)
        self.feesValueLabel = QtWidgets.QLabel(TransferDialog)
        self.feesValueLabel.setText("")
        self.feesValueLabel.setObjectName("feesValueLabel")
        self.gridLayout.addWidget(self.feesValueLabel, 6, 1, 1, 1)
        self.senderBalanceValueLabel = QtWidgets.QLabel(TransferDialog)
        self.senderBalanceValueLabel.setText("")
        self.senderBalanceValueLabel.setObjectName("senderBalanceValueLabel")
        self.gridLayout.addWidget(self.senderBalanceValueLabel, 0, 8, 1, 1)
        self.senderNameOrAddressValueLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.senderNameOrAddressValueLabel.setFont(font)
        self.senderNameOrAddressValueLabel.setText("")
        self.senderNameOrAddressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.senderNameOrAddressValueLabel.setObjectName("senderNameOrAddressValueLabel")
        self.gridLayout.addWidget(self.senderNameOrAddressValueLabel, 1, 1, 1, 4)
        self.loaderIconLabel = QtWidgets.QLabel(TransferDialog)
        self.loaderIconLabel.setEnabled(True)
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.gridLayout.addWidget(self.loaderIconLabel, 6, 4, 1, 1)
        self.commentLabel = QtWidgets.QLabel(TransferDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.commentLabel.setFont(font)
        self.commentLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.commentLabel.setObjectName("commentLabel")
        self.gridLayout.addWidget(self.commentLabel, 4, 0, 1, 1)
        self.commentLineEdit = QtWidgets.QLineEdit(TransferDialog)
        self.commentLineEdit.setText("")
        self.commentLineEdit.setObjectName("commentLineEdit")
        self.gridLayout.addWidget(self.commentLineEdit, 4, 1, 1, 4)

        self.retranslateUi(TransferDialog)
        QtCore.QMetaObject.connectSlotsByName(TransferDialog)
        TransferDialog.setTabOrder(self.recipientNameOrAddressLineEdit, self.amountDoubleSpinBox)
        TransferDialog.setTabOrder(self.amountDoubleSpinBox, self.amountUnitComboBox)
        TransferDialog.setTabOrder(self.amountUnitComboBox, self.feesButton)
        TransferDialog.setTabOrder(self.feesButton, self.sendButton)

    def retranslateUi(self, TransferDialog):
        #  _translate = QtCore.QCoreApplication.translate
        TransferDialog.setWindowTitle(_("Transfer"))
        self.amountLabel.setText(_("Amount"))
        self.sendButton.setText(_("Send"))
        self.senderNameOrAddressLineEdit.setPlaceholderText(_("Name or address"))
        self.recipientLabel.setText(_("Recipient"))
        self.recipientNameOrAddressLineEdit.setPlaceholderText(_("Name or address"))
        self.feesButton.setText(_("Fees"))
        self.senderLabel.setText(_("Sender"))
        self.feesLabel.setText(_("Fees"))
        self.commentLabel.setText(_("Comment"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.currency import (
    NetworkCurrencyException,
    NetworkCurrencyInterface,
)


class NetworkCurrency(NetworkCurrencyInterface):
    """
    NetworkCurrency class
    """

    def get_token_decimals(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_token_decimals.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.rpc_request(
                "system_properties", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result["tokenDecimals"]  # type: ignore

    def get_token_symbol(self) -> str:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_token_symbol.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.rpc_request(
                "system_properties", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result["tokenSymbol"]  # type: ignore

    def get_universal_dividend(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_universal_dividend.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.query("UniversalDividend", "CurrentUd")
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result.value

    def get_monetary_mass(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_monetary_mass.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.query(
                "UniversalDividend", "MonetaryMass"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result.value

    def get_members_count(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_members_count.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.query(
                "Membership", "CounterForMembership"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result.value

    def get_expected_block_time(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_expected_block_time.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.get_constant(
                "Babe", "ExpectedBlockTime"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        return result.value

    def get_epoch_duration(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkCurrencyInterface.get_epoch_duration.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkCurrencyException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.get_constant("Babe", "EpochDuration")
        except Exception as exception:
            logging.exception(exception)
            raise NetworkCurrencyException(exception)

        epoch_duration_in_blocks = result.value
        block_duration_in_ms = self.get_expected_block_time()

        return epoch_duration_in_blocks * block_duration_in_ms



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from datetime import datetime, timedelta
from typing import List, Optional

from substrateinterface import Keypair
from substrateinterface.exceptions import SubstrateRequestException

from tikka.domains.entities.smith import Smith, SmithStatus
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.smiths import (
    NetworkSmithsException,
    NetworkSmithsInterface,
)


class NetworkSmiths(NetworkSmithsInterface):
    """
    NetworkSmiths class
    """

    status_map = {
        "Invited": SmithStatus.INVITED,
        "Pending": SmithStatus.PENDING,
        "Smith": SmithStatus.SMITH,
        "Excluded": SmithStatus.EXCLUDED,
    }

    def get_smith(self, identity_index: int) -> Optional[Smith]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.get_smith.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.query(
                "SmithMembers", "Smiths", [identity_index]
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        smith = None
        if result.value is not None:
            if result.value["expires_on"] is not None:
                expire_on_datetime = self.get_datetime_from_epoch(
                    result.value["expires_on"]
                )
            else:
                expire_on_datetime = None
            smith = Smith(
                identity_index=identity_index,
                status=self.status_map[result.value["status"]],
                expire_on=expire_on_datetime,
                certifications_received=result.value["received_certs"],
                certifications_issued=result.value["issued_certs"],
            )

        return smith

    def get_smiths(self, identity_indice: List[int]) -> List[Optional[Smith]]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.get_smiths.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        storage_keys = []
        for identity_index in identity_indice:
            storage_keys.append(
                self.connections.rpc.client.create_storage_key(
                    "SmithMembers", "Smiths", [identity_index]
                )
            )

        try:
            multi_result = self.connections.rpc.client.query_multi(storage_keys)
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        smiths: List[Optional[Smith]] = []
        for index, (storage_key, value_obj) in enumerate(multi_result):
            if value_obj.value is not None:
                if value_obj.value["expires_on"] is not None:
                    expire_on_datetime = self.get_datetime_from_epoch(
                        value_obj.value["expires_on"]
                    )
                else:
                    expire_on_datetime = None
                smiths.append(
                    Smith(
                        identity_index=storage_keys[index].params[0],
                        status=self.status_map[value_obj.value["status"]],
                        expire_on=expire_on_datetime,
                        certifications_received=value_obj.value["received_certs"],
                        certifications_issued=value_obj.value["issued_certs"],
                    )
                )
            else:
                smiths.append(None)

        return smiths

    def invite(self, keypair: Keypair, identity_index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.invite.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        params = {
            "receiver": identity_index,
        }

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembers",
                call_function="invite_smith",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        if result.is_success is False:
            logging.error(result.error_message)
            raise NetworkSmithsException(result.error_message["name"])

    def accept_invitation(self, keypair: Keypair) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.accept_invitation.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembers",
                call_function="accept_invitation",
                call_params=None,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        try:
            # fixme: code stuck infinitely if no blocks are created on blockchain
            #       should have a timeout option
            result = self.connections.rpc.client.submit_extrinsic(
                extrinsic, wait_for_inclusion=True
            )
            logging.debug(
                "Extrinsic '%s' sent and included in block '%s'",
                result.extrinsic_hash,
                result.block_hash,
            )
        except SubstrateRequestException as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        if result.is_success is False:
            raise NetworkSmithsException(result.error_message["name"])

    def certify(self, keypair: Keypair, identity_index: int) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkSmithsInterface.certify.__doc__
        )
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        params = {
            "receiver": identity_index,
        }

        try:
            call = self.connections.rpc.client.compose_call(
                call_module="SmithMembers",
                call_function="certify_smith",
                call_params=params,
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        try:
            extrinsic = self.connections.rpc.client.create_signed_extrinsic(
                call=call, keypair=keypair
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        # fixme: code stuck infinitely if no blocks are created on blockchain
        #       should have a timeout option
        result = self.connections.rpc.client.submit_extrinsic(
            extrinsic, wait_for_inclusion=True
        )
        logging.debug(
            "Extrinsic '%s' sent and included in block '%s'",
            result.extrinsic_hash,
            result.block_hash,
        )
        if result.is_success is False:
            logging.error(result.error_message)
            raise NetworkSmithsException(result.error_message["name"])

    def get_datetime_from_epoch(self, epoch_index: int) -> Optional[datetime]:
        """
        Return a datetime object from an epoch index

        :param epoch_index: Epoch number
        :return:
        """
        if not self.connections.is_connected() or self.connections.rpc.client is None:
            raise NetworkSmithsException(NetworkConnectionError())

        try:
            result = self.connections.rpc.client.get_constant("Babe", "EpochDuration")
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        epoch_duration_in_blocks = result.value

        try:
            result = self.connections.rpc.client.get_constant(
                "Babe", "ExpectedBlockTime"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)

        block_duration_in_ms = result.value
        epoch_duration_in_ms = epoch_duration_in_blocks * block_duration_in_ms

        try:
            current_epoch_result = self.connections.rpc.client.query(
                "Babe", "EpochIndex"
            )
        except Exception as exception:
            logging.exception(exception)
            raise NetworkSmithsException(exception)
        current_epoch_index = current_epoch_result.value

        current_time = datetime.now()
        epoch_diff = epoch_index - current_epoch_index
        if epoch_diff < 0:
            block_time = current_time - timedelta(
                milliseconds=abs(epoch_diff) * epoch_duration_in_ms
            )
        else:
            block_time = current_time + timedelta(
                milliseconds=abs(epoch_diff) * epoch_duration_in_ms
            )

        return block_time
